Class: Rightscale::HttpConnection
- Inherits:
-
Object
- Object
- Rightscale::HttpConnection
- Defined in:
- lib/right_http_connection.rb
Overview
HttpConnection maintains a persistent HTTP connection to a remote server. Each instance maintains its own unique connection to the HTTP server. HttpConnection makes a best effort to receive a proper HTTP response from the server, although it does not guarantee that this response contains a HTTP Success code.
On low-level errors (TCP/IP errors) HttpConnection invokes a reconnect and retry algorithm. Note that although each HttpConnection object has its own connection to the HTTP server, error handling is shared across all connections to a server. For example, if there are three connections to www.somehttpserver.com, a timeout error on one of those connections will cause all three connections to break and reconnect. A connection will not break and reconnect, however, unless a request becomes active on it within a certain amount of time after the error (as specified by HTTP_CONNECTION_RETRY_DELAY). An idle connection will not break even if other connections to the same server experience errors.
A HttpConnection will retry a request a certain number of times (as defined by HTTP_CONNNECTION_RETRY_COUNT). If all the retries fail, an exception is thrown and all HttpConnections associated with a server enter a probationary period defined by HTTP_CONNECTION_RETRY_DELAY. If the user makes a new request subsequent to entering probation, the request will fail immediately with the same exception thrown on probation entry. This is so that if the HTTP server has gone down, not every subsequent request must wait for a connect timeout before failing. After the probation period expires, the internal state of the HttpConnection is reset and subsequent requests have the full number of potential reconnects and retries available to them.
Constant Summary collapse
- HTTP_CONNECTION_RETRY_COUNT =
Number of times to retry the request after encountering the first error
3- HTTP_CONNECTION_OPEN_TIMEOUT =
Throw a Timeout::Error if a connection isn't established within this number of seconds
5- HTTP_CONNECTION_READ_TIMEOUT =
Throw a Timeout::Error if no data have been read on this connnection within this number of seconds
120- HTTP_CONNECTION_RETRY_DELAY =
Length of the post-error probationary period during which all requests will fail
15- DEBIAN_CA_FILE =
Location of the system CA certificates file on Debian/Ubuntu systems, as generated by the ca-certificates package.
'/etc/ssl/certs/ca-certificates.crt'- @@params =
class methods
{}
Instance Attribute Summary collapse
-
#http ⇒ Object
------------------ instance methods ------------------.
-
#logger ⇒ Object
Returns the value of attribute logger.
-
#params ⇒ Object
see @@params.
-
#server ⇒ Object
Returns the value of attribute server.
Class Method Summary collapse
-
.params ⇒ Object
Query the global (class-level) parameters:.
-
.params=(params) ⇒ Object
Set the global (class-level) parameters.
Instance Method Summary collapse
- #finish(reason = '') ⇒ Object
- #get_ca_file ⇒ Object
- #get_ca_file_from_env ⇒ Object
- #get_param(name) ⇒ Object
- #get_system_ca_file ⇒ Object
-
#initialize(params = {}) ⇒ HttpConnection
constructor
Params hash: :user_agent => 'www.HostName.com' # String to report as HTTP User agent :ca_file => 'path_to_file' # A path of a CA certification file in PEM format.
-
#local_read_size=(newsize) ⇒ Object
Set the maximum size (in bytes) of a single read from local data sources like files.
-
#local_read_size? ⇒ Boolean
Query for the maximum size (in bytes) of a single read from local data sources like files.
-
#request(request_params, &block) ⇒ Object
Send HTTP request to server.
-
#socket_read_size=(newsize) ⇒ Object
Set the maximum size (in bytes) of a single read from the underlying socket.
-
#socket_read_size? ⇒ Boolean
Query for the maximum size (in bytes) of a single read from the underlying socket.
Constructor Details
#initialize(params = {}) ⇒ HttpConnection
Params hash: :user_agent => 'www.HostName.com' # String to report as HTTP User agent :ca_file => 'path_to_file' # A path of a CA certification file in PEM format. The file can contain several CA certificates. :logger => Logger object # If omitted, HttpConnection logs to STDOUT :exception => Exception to raise # The type of exception to raise if a request repeatedly fails. RuntimeError is raised if this parameter is omitted. :http_connection_retry_count # by default == Rightscale::HttpConnection.params :http_connection_open_timeout # by default == Rightscale::HttpConnection.params :http_connection_read_timeout # by default == Rightscale::HttpConnection.params :http_connection_retry_delay # by default == Rightscale::HttpConnection.params
152 153 154 155 156 157 158 159 160 161 162 163 |
# File 'lib/right_http_connection.rb', line 152 def initialize(params={}) @params = params @params[:http_connection_retry_count] ||= @@params[:http_connection_retry_count] @params[:http_connection_open_timeout] ||= @@params[:http_connection_open_timeout] @params[:http_connection_read_timeout] ||= @@params[:http_connection_read_timeout] @params[:http_connection_retry_delay] ||= @@params[:http_connection_retry_delay] @http = nil @server = nil @logger = get_param(:logger) || (RAILS_DEFAULT_LOGGER if defined?(RAILS_DEFAULT_LOGGER)) || Logger.new(STDOUT) end |
Instance Attribute Details
#http ⇒ Object
instance methods
137 138 139 |
# File 'lib/right_http_connection.rb', line 137 def http @http end |
#logger ⇒ Object
Returns the value of attribute logger.
140 141 142 |
# File 'lib/right_http_connection.rb', line 140 def logger @logger end |
#params ⇒ Object
see @@params
139 140 141 |
# File 'lib/right_http_connection.rb', line 139 def params @params end |
#server ⇒ Object
Returns the value of attribute server.
138 139 140 |
# File 'lib/right_http_connection.rb', line 138 def server @server end |
Class Method Details
.params ⇒ Object
Query the global (class-level) parameters:
:user_agent => 'www.HostName.com' # String to report as HTTP User agent :ca_file => 'path_to_file' # Path to a CA certification file in # PEM format. The file can contain # several CA certificates. If this # parameter isn't set, # HttpConnection will check for a # system CA file. If no CA file is # found, HTTPS certs won't be # verified. # # This parameter may be overridden # with the RIGHT_HTTP_CA_FILE # environment variable. :logger => Logger object # If omitted, HttpConnection logs to STDOUT :exception => Exception to raise # The type of exception to raise # if a request repeatedly fails. RuntimeError is raised if this parameter is omitted. :http_connection_retry_count # by default == Rightscale::HttpConnection::HTTP_CONNECTION_RETRY_COUNT :http_connection_open_timeout # by default == Rightscale::HttpConnection::HTTP_CONNECTION_OPEN_TIMEOUT :http_connection_read_timeout # by default == Rightscale::HttpConnection::HTTP_CONNECTION_READ_TIMEOUT :http_connection_retry_delay # by default == Rightscale::HttpConnection::HTTP_CONNECTION_RETRY_DELAY
125 126 127 |
# File 'lib/right_http_connection.rb', line 125 def self.params @@params end |
.params=(params) ⇒ Object
Set the global (class-level) parameters
130 131 132 |
# File 'lib/right_http_connection.rb', line 130 def self.params=(params) @@params = params end |
Instance Method Details
#finish(reason = '') ⇒ Object
435 436 437 438 439 440 441 |
# File 'lib/right_http_connection.rb', line 435 def finish(reason = '') if @http && @http.started? reason = ", reason: '#{reason}'" unless reason.blank? @logger.info("Closing #{@http.use_ssl? ? 'HTTPS' : 'HTTP'} connection to #{@http.address}:#{@http.port}#{reason}") @http.finish end end |
#get_ca_file ⇒ Object
443 444 445 |
# File 'lib/right_http_connection.rb', line 443 def get_ca_file get_ca_file_from_env || get_param(:ca_file) || get_system_ca_file end |
#get_ca_file_from_env ⇒ Object
447 448 449 |
# File 'lib/right_http_connection.rb', line 447 def get_ca_file_from_env ENV['RIGHT_HTTP_CA_FILE'] end |
#get_param(name) ⇒ Object
165 166 167 |
# File 'lib/right_http_connection.rb', line 165 def get_param(name) @params[name] || @@params[name] end |
#get_system_ca_file ⇒ Object
451 452 453 454 455 456 457 458 |
# File 'lib/right_http_connection.rb', line 451 def get_system_ca_file system_ca_file = DEBIAN_CA_FILE if File.readable?(system_ca_file) system_ca_file else nil end end |
#local_read_size=(newsize) ⇒ Object
Set the maximum size (in bytes) of a single read from local data sources like files. This can be used to tune the performance of, for example, a streaming PUT of a large buffer.
193 194 195 |
# File 'lib/right_http_connection.rb', line 193 def local_read_size=(newsize) Net::HTTPGenericRequest.local_read_size=(newsize) end |
#local_read_size? ⇒ Boolean
Query for the maximum size (in bytes) of a single read from local data sources like files. This is important, for example, in a streaming PUT of a large buffer.
186 187 188 |
# File 'lib/right_http_connection.rb', line 186 def local_read_size? Net::HTTPGenericRequest.local_read_size? end |
#request(request_params, &block) ⇒ Object
Send HTTP request to server
request_params hash:
:server => 'www.HostName.com' # Hostname or IP address of HTTP server
:port => '80' # Port of HTTP server
:protocol => 'https' # http and https are supported on any port
:request => 'requeststring' # Fully-formed HTTP request to make
Raises RuntimeError, Interrupt, and params[:exception] (if specified in new).
344 345 346 347 348 349 350 351 352 353 354 355 356 357 358 359 360 361 362 363 364 365 366 367 368 369 370 371 372 373 374 375 376 377 378 379 380 381 382 383 384 385 386 387 388 389 390 391 392 393 394 395 396 397 398 399 400 401 402 403 404 405 406 407 408 409 410 411 412 413 414 415 416 417 418 419 420 421 422 423 424 425 426 427 428 429 430 431 432 433 |
# File 'lib/right_http_connection.rb', line 344 def request(request_params, &block) # We save the offset here so that if we need to retry, we can return the file pointer to its initial position mypos = get_fileptr_offset(request_params) loop do request_params[:protocol] ||= (request_params[:port] == 443 ? 'https' : 'http') # (re)open connection to server if none exists or params has changed same_server_as_before = @server == request_params[:server] && @port == request_params[:port] && @protocol == request_params[:protocol] # if we are inside a delay between retries: no requests this time! # (skip this step if the endpoint has changed) if error_count > @params[:http_connection_retry_count] && error_time + @params[:http_connection_retry_delay] > Time.now && same_server_as_before # store the message (otherwise it will be lost after error_reset and # we will raise an exception with an empty text) = @logger.warn("#{err_header} re-raising same error: #{banana_message_text} " + "-- error count: #{error_count}, error age: #{Time.now.to_i - error_time.to_i}") exception = get_param(:exception) || RuntimeError raise exception.new() end # try to connect server(if connection does not exist) and get response data begin request = request_params[:request] request['User-Agent'] = get_param(:user_agent) || '' unless @http && @http.started? && same_server_as_before start(request_params) end # Detect if the body is a streamable object like a file or socket. If so, stream that # bad boy. setup_streaming(request) response = @http.request(request, &block) error_reset eof_reset return response # We treat EOF errors and the timeout/network errors differently. Both # are tracked in different statistics blocks. Note below that EOF # errors will sleep for a certain (exponentially increasing) period. # Other errors don't sleep because there is already an inherent delay # in them; connect and read timeouts (for example) have already # 'slept'. It is still not clear which way we should treat errors # like RST and resolution failures. For now, there is no additional # delay for these errors although this may change in the future. # EOFError means the server closed the connection on us. rescue EOFError => e @logger.debug("#{err_header} server #{@server} closed connection") @http = nil # if we have waited long enough - raise an exception... if raise_on_eof_exception? exception = get_param(:exception) || RuntimeError @logger.warn("#{err_header} raising #{exception} due to permanent EOF being received from #{@server}, error age: #{Time.now.to_i - eof_time.to_i}") raise exception.new("Permanent EOF is being received from #{@server}.") else # ... else just sleep a bit before new retry sleep(add_eof) # We will be retrying the request, so reset the file pointer reset_fileptr_offset(request, mypos) end rescue Exception => e # See comment at bottom for the list of errors seen... @http = nil # if ctrl+c is pressed - we have to reraise exception to terminate proggy if e.is_a?(Interrupt) && !( e.is_a?(Errno::ETIMEDOUT) || e.is_a?(Timeout::Error)) @logger.debug( "#{err_header} request to server #{@server} interrupted by ctrl-c") raise elsif e.is_a?(ArgumentError) && e..include?('wrong number of arguments (5 for 4)') # seems our net_fix patch was overriden... exception = get_param(:exception) || RuntimeError raise exception.new('incompatible Net::HTTP monkey-patch') end # oops - we got a banana: log it error_add(e.) @logger.warn("#{err_header} request failure count: #{error_count}, exception: #{e.inspect}") # We will be retrying the request, so reset the file pointer reset_fileptr_offset(request, mypos) end end end |
#socket_read_size=(newsize) ⇒ Object
Set the maximum size (in bytes) of a single read from the underlying socket. For bulk transfer, especially over fast links, this is value is critical to performance.
179 180 181 |
# File 'lib/right_http_connection.rb', line 179 def socket_read_size=(newsize) Net::BufferedIO.socket_read_size=(newsize) end |
#socket_read_size? ⇒ Boolean
Query for the maximum size (in bytes) of a single read from the underlying socket. For bulk transfer, especially over fast links, this is value is critical to performance.
172 173 174 |
# File 'lib/right_http_connection.rb', line 172 def socket_read_size? Net::BufferedIO.socket_read_size? end |